@@ -17,4 +17,4 @@ index.add_command(nodes)
 index.add_command(ursula)
 # index.add_command(ethereum)
 index.add_command(namespaces)
-# index.add_command(porter)
+index.add_command(porter)
