@@ -74,15 +74,18 @@ def add(host_address, login_name, key_path, ssh_port, nickname, namespace, netwo
 
 
 @cli.command('copy')
-@click.option('--nickname', help="A nickname to remember this host by", type=click.STRING, required=True)
-@click.option('--namespace', help="Namespace for these operations.  Used to address hosts and data locally and name hosts on cloud platforms.", type=click.STRING, default='nucypher')
-@click.option('--network', help="The Nucypher network name these hosts will run on. (default mainnet)", type=click.STRING, default=DEFAULT_NETWORK)
+@click.option('--to-namespace', help="Namespace for these operations.  Used to address hosts and data locally and name hosts on cloud platforms.", type=click.STRING, default='nucypher')
+@click.option('--to-network', help="The Nucypher network name these hosts will run on. (default mainnet)", type=click.STRING, default=DEFAULT_NETWORK)
 @click.option('--from', 'from_path', help="The 'path' of a node in /network/namespace/name format ie. '/mainnet/aws-us-east-nodes/aws-1'", type=click.STRING, required=True)
-def copy(from_path, from_name, to_network, namespace, nickname):
-    """Adds an existing node to the local config for future management."""
+def copy(from_path, to_network, to_namespace):
+    """
+        Copy a node from one namespace to another
+       
+        ie.  nucypher-ops nodes copy --to-namespace allmynodes --from mainnet/eu-central-1/europe-nodes-1
+    """
 
     try:
-        network, namespace, host_name = from_path.split('/')
+        network, namespace, host_name = from_path.lstrip('/').rstrip('/').split('/')
     except Exception as e:
         emitter.echo("please supply --from in the format of /network/namespace/node_nickname")
         return
@@ -90,16 +93,14 @@ def copy(from_path, from_name, to_network, namespace, nickname):
 
     source = CloudDeployers.get_deployer('generic')(emitter, namespace=namespace, network=network)
     host_data = source.get_host_by_name(host_name)
-    print(host_data)
     
-
-    # deployer = CloudDeployers.get_deployer('generic')(
-    #     emitter, None, None, namespace=namespace, network=network, action='add')
-    # deployer.create_nodes([name], host_address, login_name, key_path, ssh_port)
+    deployer = CloudDeployers.get_deployer('generic')(
+        emitter, namespace=to_namespace, network=to_network, action='copy')
+    deployer.add_already_configured_node(host_data)
 
 
 @cli.command('list')
-@click.option('--json', 'as_json', help="list node data", default=False, is_flag=True)
+@click.option('--json', 'as_json', help="output json", default=False, is_flag=True)
 @click.option('--all', help="list all nodes under all networks and namespaces", default=False, is_flag=True)
 @click.option('--network', help="The network whose hosts you want to see.", type=click.STRING, default=DEFAULT_NETWORK)
 @click.option('--namespace', help="The network whose hosts you want to see.", type=click.STRING, default=DEFAULT_NAMESPACE)
@@ -118,9 +119,9 @@ def list(network, namespace, all, as_json):
     
     for deployer in deployers:
         if not as_json:
-            emitter.echo(f'{network}')
+            emitter.echo(f'{deployer.network}')
         for ns, hosts in deployer.get_namespace_data(namespace=namespace):
-            if not as_json:
+            if not as_json and hosts:
                 emitter.echo(f'\t{ns}')
             for name, data in hosts:
                 if not as_json:
