@@ -24,12 +24,13 @@ import maya
 import os
 import requests
 import time
+import warnings
 from ansible.executor.playbook_executor import PlaybookExecutor
 from ansible.inventory.manager import InventoryManager
 from ansible.parsing.dataloader import DataLoader
 
 from ansible.vars.manager import VariableManager
-from base64 import b64encode
+from base64 import b64encode, b64decode
 from mako.template import Template
 from pathlib import Path
 
@@ -42,6 +43,7 @@ from nucypher_ops.constants import (
 )
 
 from nucypher_ops.ops import keygen
+from nucypher_ops.ops.contracts import NuCypherContractRegistry
 
 try:
     import boto3
@@ -49,11 +51,31 @@ except ImportError:
     pass
 
 NODE_CONFIG_STORAGE_KEY = 'configs'
+                            
+
+def needs_provider(method):
+    def inner(self, *args, **kwargs):
+        provider = self.get_local_blockchain_provider()
+        try:
+            import web3
+        except ImportError:
+            self.emitter.echo("web3 must be installed to use this functionality ('pip install web3')")
+        w3 = web3.Web3(web3.Web3.HTTPProvider(provider))
+        return method(self, w3, *args, **kwargs)
+    return inner
+
+def needs_registry(method):
+    def inner(self, *args, **kwargs):
+        if self.contract_registry is None:
+            registry = NuCypherContractRegistry(network_name=self.network)
+            self.contract_registry = {name: (address, abi) for name, version, address, abi in registry.fetch_latest_publication()} 
+        return method(self, self.contract_registry, *args, **kwargs)
+    return inner
 
 
 class BaseCloudNodeConfigurator:
 
-    NAMESSPACE_CREATE_ACTIONS = ['add', 'create']
+    NAMESSPACE_CREATE_ACTIONS = ['add', 'create', 'copy']
     application = 'ursula'
     required_fields = [
         'eth_provider',
@@ -96,6 +118,7 @@ class BaseCloudNodeConfigurator:
         self.action = action
         self.resource_name = resource_name
         self.kwargs = kwargs
+        self.contract_registry = None
 
         self.envvars = envvars or []
         if self.envvars:
@@ -128,8 +151,13 @@ class BaseCloudNodeConfigurator:
             self.namespace / self.config_filename
         self.config_dir = self.config_path.parent
 
+        # print (self.config_path)
         if self.config_path.exists():
-            self.config = json.load(open(self.config_path))
+            try:
+                self.config = json.load(open(self.config_path))
+            except json.decoder.JSONDecodeError as e:
+                self.emitter.echo(f"could not decode config file at: {self.config_path}")
+                raise
             self.namespace_network = self.config['namespace']
         elif kwargs.get('read_only'):
             self.config = {
@@ -201,6 +229,10 @@ class BaseCloudNodeConfigurator:
     def _provider_deploy_attrs(self):
         return []
 
+    @property
+    def has_wallet(self):
+        return self.config.get('local_wallet_keystore') is not None
+
     def _configure_provider_params(self):
         pass
 
@@ -460,29 +492,29 @@ class BaseCloudNodeConfigurator:
         self.update_captured_instance_data(self.output_capture)
         self.give_helpful_hints(node_names, backup=True, playbook=playbook)
 
-    def get_worker_status(self, node_names):
+    def get_worker_status(self, node_names, fast=False):
 
         playbook = Path(PLAYBOOKS).joinpath('get_workers_status.yml')
+        if not fast:
+            self.update_generate_inventory(node_names)
 
-        self.update_generate_inventory(node_names)
+            loader = DataLoader()
+            inventory = InventoryManager(
+                loader=loader, sources=self.inventory_path)
+            callback = AnsiblePlayBookResultsCollector(sock=self.emitter, return_results=self.output_capture, filter_output=[
+                                                    "Print Ursula Status Data", "Print Last Log Line"])
+            variable_manager = VariableManager(loader=loader, inventory=inventory)
 
-        loader = DataLoader()
-        inventory = InventoryManager(
-            loader=loader, sources=self.inventory_path)
-        callback = AnsiblePlayBookResultsCollector(sock=self.emitter, return_results=self.output_capture, filter_output=[
-                                                   "Print Ursula Status Data", "Print Last Log Line"])
-        variable_manager = VariableManager(loader=loader, inventory=inventory)
-
-        executor = PlaybookExecutor(
-            playbooks=[playbook],
-            inventory=inventory,
-            variable_manager=variable_manager,
-            loader=loader,
-            passwords=dict(),
-        )
-        executor._tqm._stdout_callback = callback
-        executor.run()
-        self.update_captured_instance_data(self.output_capture)
+            executor = PlaybookExecutor(
+                playbooks=[playbook],
+                inventory=inventory,
+                variable_manager=variable_manager,
+                loader=loader,
+                passwords=dict(),
+            )
+            executor._tqm._stdout_callback = callback
+            executor.run()
+            self.update_captured_instance_data(self.output_capture)
 
         self.give_helpful_hints(node_names, playbook=playbook)
 
@@ -610,10 +642,20 @@ class BaseCloudNodeConfigurator:
             yield (ns, dep.get_all_hosts())
 
     def get_host_by_name(self, host_name):
-        return next([host_data for node_name, host_data in self.get_all_hosts() if node_name == host_name])
+        try:
+            return next(host_data for node_name, host_data in self.get_all_hosts() if node_name == host_name)
+        except StopIteration:
+            return None
 
     def get_all_hosts(self):
-        return [(node_name, host_data) for node_name, host_data in self.config['instances'].items()]
+        return [(node_name, host_data) for node_name, host_data in self.config.get('instances', {}).items()]
+
+    def add_already_configured_node(self, host_data):
+        if self.get_host_by_name(host_data['host_nickname']):
+            raise AttributeError(f"Host with nickname {host_data['host_nickname']} already exists in {self.network}/{self.namespace}")
+        host_data['index'] = self.instance_count
+        self.config['instances'][host_data['host_nickname']] = host_data
+        self._write_config()
 
     def destroy_resources(self, node_names):
         node_names = [s for s in node_names if s in [
@@ -652,29 +694,64 @@ class BaseCloudNodeConfigurator:
             self.emitter.echo(
                 f" keypair file: {self.config['keypair_path']}", color='yellow')
 
-        if nodes := [h for h in self.get_all_hosts() if h[0] in node_names]:
-            self.emitter.echo("Host Info")
-            for node_name, host_data in nodes:
-                dep = CloudDeployers.get_deployer(host_data['provider'])(
-                    self.emitter,
-                    pre_config=self.config,
-                    namespace=self.namespace,
-                    network=self.network
-                )
-                self.emitter.echo(
-                    f"\t{node_name}: {host_data['publicaddress']}")
-                self.emitter.echo(
-                    f"\t\t ssh: {dep.format_ssh_cmd(host_data)}", color="yellow")
-                if host_data.get('operator address'):
-                    self.emitter.echo(
-                        f"\t\t operator address: {host_data['operator address']}")
-
         if backup:
             self.emitter.echo(
                 " *** Local backups containing sensitive data may have been created. ***", color="red")
             self.emitter.echo(
                 f" Backup data can be found here: {self.config_dir}/remote_worker_backups/")
 
+        if nodes := [h for h in self.get_all_hosts() if h[0] in node_names]:
+            self.emitter.echo("Host Info")
+            for node_name, host_data in nodes:
+                self.print_node_data(node_name, host_data)
+    
+    def print_node_data(self, node_name, host_data):
+        warnings.filterwarnings("ignore")
+        dep = CloudDeployers.get_deployer(host_data['provider'])(
+            self.emitter,
+            pre_config=self.config,
+            namespace=self.namespace,
+            network=self.network
+        )
+        self.emitter.echo(
+            f"\t{node_name}: {host_data['publicaddress']}")
+        self.emitter.echo(
+            f"\t\t {dep.format_ssh_cmd(host_data)}", color="yellow")
+        if host_data.get('operator address'):
+            self.emitter.echo(
+                f"\t\t operator address: {host_data['operator address']}")
+            if self.config.get('local_blockchain_provider'):
+                wallet_balance = self.get_wallet_balance(host_data['operator address'], eth=True)
+                self.emitter.echo(
+                    f"\t\t operator ETH balance: {wallet_balance}"
+                )
+                staking_provider = None
+                try:
+                    staking_provider = self.get_staking_provider(host_data['operator address'])
+                    self.emitter.echo(f"\t\t staking provider address: {staking_provider}")
+                    if staking_provider:
+                        # if we have a staking provider, lets check if the node is confirmed
+                        is_confirmed = self.check_is_confirmed(host_data['operator address'])
+                        self.emitter.echo(f"\t\t operator confirmed: {is_confirmed}")
+                        stake_amount = self.get_stake_amount(staking_provider)
+                        self.emitter.echo(f"\t\t staked amount: {stake_amount:,}")
+                        if is_confirmed:
+                            # if the node is confirmed, we should be able to query it
+                            try:
+                                node_response = self.query_active_node(host_data['publicaddress'])
+                                self.emitter.echo("\t\t active node status:")
+                                self.emitter.echo(f"\t\t\tnickname: {node_response['nickname']['text']}")
+                                self.emitter.echo(f"\t\t\trest url: {node_response['rest_url']}")
+                                self.emitter.echo(f"\t\t\tknown nodes: {len(node_response['known_nodes'])}")
+                                self.emitter.echo(f"\t\t\tfleet state: {len(node_response['fleet_state'])}")
+                            except Exception as e:
+                                print (e)
+                except Exception as e:
+                    raise e
+                if not staking_provider:
+                    self.emitter.echo(f"\t\t staking provider: NOT BOUND TO STAKING PROVIDER")
+
+
     def format_ssh_cmd(self, host_data):
 
         keypair = ''
@@ -697,6 +774,11 @@ class BaseCloudNodeConfigurator:
         self.emitter.echo(
             "Back this up and keep it safe to enable restoration of Ursula nodes in the event of hard disk failure or other data loss.", color="red")
 
+    def new_mnemonic(self):
+        wallet = keygen.generate()
+        self.config['keystoremnemonic'] = wallet.mnemonic()
+        self.alert_new_mnemonic(wallet)
+
     def migrate_5_6(self):
         for index, instance in enumerate(self.config['instances'].keys()):
             if not self.config['instances'][instance].get('index'):
@@ -709,11 +791,8 @@ class BaseCloudNodeConfigurator:
         if self.config.get('keyringpassword'):
             self.config['keystorepassword'] = self.config.get(
                 'keyringpassword')
-
         if not self.config.get('keystoremnemonic'):
-            wallet = keygen.generate()
-            self.config['keystoremnemonic'] = wallet.mnemonic()
-            self.alert_new_mnemonic(wallet)
+            self.new_mnemonic()       
         self._write_config()
 
     def migrate(self, current=5, target=6):
@@ -736,6 +815,94 @@ class BaseCloudNodeConfigurator:
                     del self.config['instances'][node_name]
                     self._write_config()
 
+    def get_local_blockchain_provider(self):
+        if not self.config.get('local_blockchain_provider'):
+            blockchain_provider = self.emitter.prompt("Please enter a blockchain provider for this local wallet to access the blockchain.")
+            self.config['local_blockchain_provider'] = blockchain_provider
+            self._write_config()
+
+        return self.config.get('local_blockchain_provider')
+
+    def get_or_create_local_wallet(self, password):
+
+        try:
+            import web3
+        except ImportError:
+            self.emitter.echo("web3 must be installed to use this functionality ('pip install web3')")
+
+        if not self.has_wallet:
+            account = web3.Account.create() # uses index 0 by default which will not be used by any subsequent node
+            keystore = web3.Account.encrypt(account.privateKey, password)
+            keystore_b64 = b64encode(json.dumps(keystore).encode()).decode('utf-8')
+            self.config['local_wallet_keystore'] = keystore_b64
+            self._write_config()
+
+        account_keystore = b64decode(self.config['local_wallet_keystore']).decode()
+        return web3.Account.from_key(web3.Account.decrypt(account_keystore, password))
+
+    @needs_provider
+    def get_wallet_balance(self, web3, address, eth=False):        
+        balance =  web3.eth.get_balance(address)
+        if eth:
+            return web3.fromWei(balance, 'ether')
+        return balance
+    
+    @needs_registry
+    @needs_provider
+    def get_staking_provider(self, web3, contracts, address):        
+        contract_address, abi =  contracts['SimplePREApplication']
+        contract = web3.eth.contract(abi=abi, address=contract_address)
+        return contract.functions.stakingProviderFromOperator(address).call()
+
+    @needs_registry
+    @needs_provider
+    def check_is_confirmed(self, web3, contracts, address):        
+        contract_address, abi =  contracts['SimplePREApplication']
+        contract = web3.eth.contract(abi=abi, address=contract_address)
+        return contract.functions.isOperatorConfirmed(address).call()
+
+    @needs_registry
+    @needs_provider
+    def get_stake_amount(self, web3, contracts, address):        
+        contract_address, abi =  contracts['SimplePREApplication']
+        contract = web3.eth.contract(abi=abi, address=contract_address)
+        balance = contract.functions.authorizedStake(address).call()
+        return int(web3.fromWei(balance, 'ether'))
+
+    def query_active_node(self, public_address):
+        return requests.get(f"https://{public_address}:9151/status/?json=true", verify=False).json()
+
+    @needs_provider
+    def fund_nodes(self, web3, wallet, node_names, amount):
+        hosts = [h for h in self.get_all_hosts() if h[0] in node_names]
+        for name, host in hosts:
+            if not host.get('operator address'):
+                raise AttributeError(f"missing operator address for node: {host['host_nickname']}.  Deploy ursula first to create an operator address.")
+
+            host_op_address = host.get('operator address')
+            balance =  web3.eth.get_balance(host_op_address)
+            existing_balance = self.get_wallet_balance(host_op_address, eth=True)
+            self.emitter.echo(f"existing balance for node: {name}: {existing_balance}")
+            if existing_balance >= amount:
+                self.emitter.echo(f"host {name} already has {existing_balance} ETH.  funded.")
+            else:
+                tx_hash = self.send_eth(wallet, host_op_address, amount)
+                self.emitter.echo(f"Broadcast transaction {tx_hash} for node: {host['host_nickname']}")
+
+    @needs_provider
+    def send_eth(self, web3, wallet, destination_address, amount_eth):
+
+        transaction = {
+                    "nonce": web3.eth.getTransactionCount(wallet.address, 'pending'),
+                    "from": wallet.address,
+                    "to": destination_address,
+                    "value": web3.toWei(amount_eth, 'ether'),
+                    "gas": 21000,
+                    "gasPrice": web3.eth.gasPrice * 2
+                }
+        signed_tx = wallet.sign_transaction(transaction)
+        return web3.eth.send_raw_transaction(signed_tx.rawTransaction).hex()
+        
 
 class DigitalOceanConfigurator(BaseCloudNodeConfigurator):
 
@@ -902,6 +1069,7 @@ class AWSNodeConfigurator(BaseCloudNodeConfigurator):
         'eu-central-1': 'ami-0c960b947cbb2dd16',  # Frankfurt
         'ap-northeast-1': 'ami-09b86f9709b3c33d4',  # Tokyo
         'ap-southeast-1': 'ami-093da183b859d5a4b',  # Singapore
+        'sa-east-1': 'ami-090006f29ecb2d79a'
     }
 
     preferred_platform = 'ubuntu-focal'  # unused
@@ -1199,8 +1367,11 @@ class AWSNodeConfigurator(BaseCloudNodeConfigurator):
 
     def create_new_node(self, node_name):
 
+        if not self.EC2_AMI_LOOKUP.get(self.AWS_REGION) or os.environ.get('NUCYPHER_OPS_AWS_AMI'):
+            raise AttributeError("Sorry nucypher-ops does not automatically support this region.  Please specify an ami for your instances by setting envar `export NUCYPHER_OPS_AWS_AMI=ami-xxxxxxxxxxxxxx`")
+
         params = dict(
-            ImageId=self.EC2_AMI_LOOKUP.get(self.AWS_REGION),
+            ImageId=os.environ.get('NUCYPHER_OPS_AWS_AMI') or self.EC2_AMI_LOOKUP.get(self.AWS_REGION),
             InstanceType=self.EC2_INSTANCE_SIZE,
             KeyName=self.keypair,
             NetworkInterfaces=[
@@ -1251,6 +1422,7 @@ class GenericConfigurator(BaseCloudNodeConfigurator):
     provider_name = 'generic'
 
     def _write_config(self):
+
         if not self.config_path.exists() and self.action not in self.NAMESSPACE_CREATE_ACTIONS:
             raise AttributeError(
                 f"Namespace/config '{self.namespace}' does not exist. Show existing namespaces: `nucypher nodes list-namespaces` or create a namespace: `nucypher nodes create`")
@@ -1414,3 +1586,4 @@ class CloudDeployers:
     @staticmethod
     def get_deployer(name):
         return getattr(CloudDeployers, name)
+
